﻿Imports System
Imports System.Collections.Generic
Imports System.Linq
Imports System.Text
Imports System.IO
Imports Microsoft.AnalysisServices.AdomdClient
Imports System.Runtime.Serialization.Formatters.Binary

Namespace Tests
	Public Class BaseMDXTest
		Private connectionstring As String
		Private baseOutputDir As String
		Private outputDir As String
		Private recreateTemplate As Boolean = False
		Private Shared testReportHtml As String
		Private Shared numPassed As Integer = 0
		Private Shared numFailed As Integer = 0

'''         Construct BaseMDX Test
'''         * connectionstring - The URL of the Analysis Server on which the MDX query is executed
'''         * baseOutputDir - The output directory under which the MDX query results are stored
'''         * recreateTemplate - Force the recreation of the MDX query result template
'''         *
		Public Sub New(ByVal connectionstring As String, ByVal baseOutputDir As String, Optional ByVal recreateTemplate As Boolean = False)
			Me.connectionstring = connectionstring
			Me.baseOutputDir = baseOutputDir
			Me.recreateTemplate = recreateTemplate
			If testReportHtml Is Nothing Then
				testReportHtml = "<html><h1>Test Report</h1><table border=""1""><tr><th>Test Case</th><th>Pass/Fail</th></tr>"
			End If
		End Sub

		''' Open a connection to the Analysis Services database *
		Private Shared Function GetAdomdConnection(ByVal connectionstring As String) As AdomdConnection
			Dim connection = New AdomdConnection(connectionstring)
			connection.Open()
			Return connection
		End Function

'''         Runs the given MDX query, creates a template if it doesn't exist, and stores the results 
'''         * mdxTestTitle - The name of the unit-test
'''         * mdxQuery - The MDX query to test
'''         *
		Public Function TestMDXQuery(ByVal mdxTestTitle As String, ByVal mdxQuery As String) As Boolean
			Dim templateCubeDict As New Dictionary(Of String, CubeRow)()
            Dim templateHtml As String
            Dim resultHtml As String = "<html><h2>Test Title</h2><p>" & mdxTestTitle & "</p><h2>MDX Query</h2><p>" & mdxQuery & "</p>" & VBCrLf
			' If the template for the given test already exists and template recreation has not been requested
			' then run the comparison.
            If (Not recreateTemplate) AndAlso TemplateExists(mdxTestTitle) Then
                resultHtml &= "<h2>Original Template</h2>" & VBCrLf
                ' Lots the origional template HTML
                templateHtml = loadTemplate(mdxTestTitle, templateCubeDict)
                resultHtml &= templateHtml & VBCrLf
                resultHtml &= "<h2>Test Results</h2>" & VBCrLf
                ' Run the MDX query and compare against the template
                Dim result As CubeComparisonResult = RunComparison(mdxQuery, templateCubeDict)
                resultHtml &= result.getResultHtml & VBCrLf
                resultHtml += If(result.hasPassed, "<h2 stlye=""color:green"";>Test Passes!</h2>", "<h2 style=""color:red;"">Test Fails</h2>")
                resultHtml &= "</html>"
                ' Persist the MDX query result
                Dim testReportFile As String = PersistResult(mdxTestTitle, resultHtml)
                If result.hasPassed Then
                    numPassed += 1
                Else
                    numFailed += 1
                End If
                testReportHtml &= "<tr><td><a href=""file:\" & testReportFile & """>" & mdxTestTitle & "</a></td><td>" & (If(result.hasPassed, "Pass", "Fail")) & "</td></tr>"
                ' Persist the HTML summary report
                PersistTestReport(numPassed, numFailed)
                Return result.hasPassed
            Else ' Else build the MDX query result template if the template does not previously exist or a request to force the re-creation of the template.
                resultHtml &= "<h2>Original Template</h2>" & VBCrLf
                ' Build the template using the MDXQuery and store it with the Dictionary<string, CubeRow>
                templateHtml = BuildTemplate(mdxQuery, templateCubeDict)
                resultHtml &= templateHtml & VBCrLf
                resultHtml &= "<h2 style=""color:red"";>Test Fails.  No Template exists.  You must validate the output by hand</h2>"
                resultHtml &= "</html>"
                ' Persist the template.  This includes the Dictionary<string, CubeRow> and the HTML version of the MDX query result
                Dim resultHtmlFile As String = PersistTemplate(mdxTestTitle, templateHtml, templateCubeDict)
                testReportHtml &= "<tr><td><a href=""file:\" & resultHtmlFile & """>" & mdxTestTitle & "</a></td><td>Fail - Template Did not Exist</td></tr>"
                ' Creation of the template is always considered a failure since the unit-test developer must hand validate
                ' the results before the template is considered usable.
                numFailed += 1
                ' Persist the HTML summary report
                PersistTestReport(numPassed, numFailed)
                Return False
            End If
        End Function

		''' Save a summary of the test results *
		Private Sub PersistTestReport(ByVal numPassed As Integer, ByVal numFailed As Integer)
			Dim testReportFile As String = baseOutputDir & "\TestReport.html"
			Dim tempReport As String = testReportHtml
			tempReport &= "</table>"
			tempReport &= "<h3>Num Passed: " & numPassed & "</h3>"
			tempReport &= "<h3>Num Failed: " & numFailed & "</h3>"
			tempReport &= "</html>"
			Using sw As StreamWriter = File.CreateText(testReportFile)
				sw.WriteLine(tempReport)
			End Using
		End Sub

		''' Load both the HTML and Object version of the MDX query result template *
		Private Function loadTemplate(ByVal mdxTestTitle As String, ByVal templateCubeDict As Dictionary(Of String, CubeRow)) As String
			Dim templateFile As String = GetTemplateDir(mdxTestTitle) & "\template.html"
            Dim templateHtml As String
			Using sr As StreamReader = File.OpenText(templateFile)
                templateHtml = sr.ReadToEnd
			End Using

			Dim cubeDictFile As String = GetTemplateDir(mdxTestTitle) & "\CubeDict"
			Using stream As Stream = File.Open(cubeDictFile, FileMode.Open)
                Dim binaryFormatter As BinaryFormatter
				Dim templateDict As Dictionary(Of String, CubeRow) = CType(binaryFormatter.Deserialize(stream), Dictionary(Of String, CubeRow))
				For Each cubeRow In templateDict
					templateCubeDict.Add(cubeRow.Key, cubeRow.Value)
				Next cubeRow
			End Using
			Return templateHtml
		End Function

		''' Save the HTML file that contains the MDX query comparison results *
		Private Function PersistResult(ByVal mdxTestTitle As String, ByVal resultHtml As String) As String
			Dim outputDir As String = baseOutputDir & "\result\" & mdxTestTitle
			If Not System.IO.Directory.Exists(outputDir) Then
				System.IO.Directory.CreateDirectory(outputDir)
			End If

			Dim resultHtmlFile As String = outputDir & "\result.html"
			Using sw As StreamWriter = File.CreateText(resultHtmlFile)
				sw.WriteLine(resultHtml)
			End Using
			Return resultHtmlFile
		End Function

'''         Save the HTML file and Dictionary<string, CubeRow> Object that contains the initial version of the
'''         * MDX query results.
'''         * mdxTestTitle - The name of the unit-test
'''         * resultHTML - The HTML representation of the MDX query results.
'''         * templateCubeDict - THe object representation of the MDX query results.
'''         *
		Private Function PersistTemplate(ByVal mdxTestTitle As String, ByVal resultHtml As String, ByVal templateCubeDict As Dictionary(Of String, CubeRow)) As String
			Dim templateDir As String = GetTemplateDir(mdxTestTitle)
			If Not System.IO.Directory.Exists(templateDir) Then
				System.IO.Directory.CreateDirectory(templateDir)
			End If

			Dim resultHtmlFile As String = templateDir & "\template.html"
			Using sw As StreamWriter = File.CreateText(resultHtmlFile)
				sw.WriteLine(resultHtml)
			End Using

			Dim cubeDictFile As String = templateDir & "\CubeDict"
			Using stream As Stream = File.Open(cubeDictFile, FileMode.Create)
                Dim formatter As BinaryFormatter
				formatter.Serialize(stream, templateCubeDict)
			End Using
			Return resultHtmlFile
		End Function

'''         
'''         * Compares the results of a MDX query with the previously peristed results
'''         *
		Private Function RunComparison(ByVal mdxQuery As String, ByVal templateCubeDict As Dictionary(Of String, CubeRow)) As CubeComparisonResult
			Dim tableResult As String = "<table border=""1"">"
			Dim testPass As Boolean = True
			' Open a connection to the Analysis server
			Using connection = GetAdomdConnection(connectionstring)
			' Execute the MDX query
			Using command = New AdomdCommand(mdxQuery, connection)
				' Get the MDX query results as a CellSet object
                    Dim cst As CellSet = command.ExecuteCellSet

				' Make sure that the result contains at least two axes.  Columns and rows.
				Dim axesCount As Integer = cst.Axes.Count
				If axesCount <> 2 Then
					Throw New Exception("The MDX Query did not return two axes.  Please check your query.")
				End If

				' Make sure that data is contained within the two axes.
                    For count = 0 To axesCount - 1
                        If cst.Axes(count).Positions.Count = 0 Then
                            Throw New Exception("The MDX query return no data.  Please check your query.")
                        End If
                    Next count

				' Get the number of column dimensions
				Dim colDimCount As Integer = cst.Axes(0).Positions(0).Members.Count
				' Get the number of row dimensions
				Dim rowDimCount As Integer = 0
				If cst.Axes(1).Positions(0).Members.Count > 0 Then
					rowDimCount = cst.Axes(1).Positions(0).Members.Count
				End If

				' The row count is the total number of Position(s) on the row axis plus the number of column dimensions
				' The reason the column dimension count is included is that the column headers themselves make up rows
				Dim rowCount As Integer = cst.Axes(1).Positions.Count + colDimCount
				' The column count is the total number of Position(s) on the column axis plus the number of row dimensions.
				' The reason the row dimension count is included is that the row headers themselves make up columns
                    Dim colCount As Integer = cst.Axes(0).Positions.Count + rowDimCount

				' Create an array of string(s) that hold the column headers
				Dim columnHeader(cst.Axes(0).Positions.Count - 1) As String

				' Create a point to hold the current CubeRow dictionary
				Dim currentCubeDict As Dictionary(Of String, CubeRow) = templateCubeDict
				' Create a pointer to hold the current CubeRow object
				Dim currentCubeRow As CubeRow = Nothing
				' Loop through all the rows within the MDX result
                    For currentRow = 0 To rowCount - 1
                        ' Create a new HTML table row
                        tableResult &= "<tr>"

                        ' If the current row is greater than the number of column dimensions then 
                        ' processing of actual row data is occurring.  Set the CubeRow pointer to a new
                        ' CubeRow and set the CubeRow Dictionary pointer to the original Dictionary passed in.
                        ' This is also the head of the Dictionary structure.
                        If currentRow >= colDimCount Then
                            currentCubeRow = New CubeRow()
                            currentCubeDict = templateCubeDict
                        End If

                        ' Loop through all the columns within the MDX result
                        For currentCol = 0 To colCount - 1
                            ' Check if the current row is dealing with column dimension captions
                            If currentRow < colDimCount Then
                                If currentCol < rowDimCount Then
                                    ' If the current column is has not reached the row dimension headers create a blank HTML cell
                                    ' These cells make up the uper left hand corner of the HTML table as place holders before the 
                                    ' HTML column headers and row headers meet.
                                    tableResult &= "<th> </th>"
                                Else
                                    ' A column header is being processed.  The column header may be many levels deep depending on
                                    ' how many dimensions are included in the column portion of the MDX query.  The location of the column header is 
                                    ' computed by taking the current column count minus the number of row dimensions since getting the total number of 
                                    ' columns involved adding the total number of row dimensions.

                                    ' Get the current column index which is the current column count - the number of row dimensions since
                                    ' the row dimensions captions are part of the columns.
                                    Dim columnIndex As Integer = currentCol - rowDimCount

                                    ' If the column header at the given location does not exist add it.
                                    If columnHeader(columnIndex) Is Nothing Then
                                        columnHeader(columnIndex) = cst.Axes(0).Positions(columnIndex).Members(currentRow).Caption
                                    Else
                                        ' Otherwise if the column header does exist append to it.
                                        columnHeader(columnIndex) &= ":" & cst.Axes(0).Positions(columnIndex).Members(currentRow).Caption
                                    End If

                                    ' Create an HTML table header cell to hold the column name and add it to the HTML template.
                                    tableResult &= "<th>" & cst.Axes(0).Positions(columnIndex).Members(currentRow).Caption & "</th>"
                                End If
                            Else
                                ' Each cell of the HTML table will be colored green (to indicate a match with the template) or red (to indicate a mismatch)
                                ' Always the optomist the cell color is initialized as green.
                                Dim color As String = "green"
                                If currentCol < rowDimCount Then
                                    ' Get the current row index which is the current row count - the number of column dimensions since the 
                                    ' column dimension captions are part of the rows.
                                    Dim rowIndex As Integer = currentRow - colDimCount
                                    ' Capture the row caption since processing at this point is not within the row values
                                    Dim rowCaption As String = cst.Axes(1).Positions(rowIndex).Members(currentCol).Caption

                                    If Not currentCubeDict.ContainsKey(rowCaption) Then
                                        ' If the template does not contain the row caption fail the test and set the color to red.
                                        ' Color all cells within this row red since this is extra data that is not in the template.
                                        color = "red"
                                        testPass = False
                                    Else
                                        ' If the template contains the row caption extract the CubeRow dictionary that
                                        ' applies to this row caption and point the current CubeRow and CubeRow Dictionary 
                                        ' to the extracted CubeRow.
                                        currentCubeRow = currentCubeDict(rowCaption)
                                        currentCubeDict = currentCubeRow.GetCubeDict
                                    End If
                                    ' Add the row caption to the HTML table appropriately colored.
                                    tableResult &= "<th bgcolor=""" & color & """>" & rowCaption & "</th>"
                                Else
                                    ' Get the current column index which is the current column count - the number of row dimensions since
                                    ' the row dimensions captions are part of the columns.
                                    Dim columnIndex As Integer = currentCol - rowDimCount
                                    ' Get the current row index which is the current row count - the number of column dimensions since the 
                                    ' column dimension captions are part of the rows.
                                    Dim rowIndex As Integer = currentRow - colDimCount
                                    ' Get the current column name for the array of column headers
                                    Dim columnName As String = columnHeader(columnIndex)
                                    ' Get the current column value based on the current column and row
                                    Dim value As String = cst(columnIndex, rowIndex).FormattedValue
                                    'if (value.Equals("621369"))
                                    'value = "621370";
                                    If currentCubeRow Is Nothing Then
                                        ' if the current cube row does not exist color all cells red and fail the test.
                                        color = "red"
                                        testPass = False
                                    Else
                                        ' Get the CubeTuple for the the current column name
                                        Dim templateValue As CubeTuple = currentCubeRow.GetCubeTuple(columnName)
                                        If templateValue Is Nothing Then
                                            ' If the column name returned by the MDX query does not exist in the template.
                                            ' then the test is considered to have failed.
                                            color = "red"
                                            testPass = False
                                        ElseIf Not templateValue.getColumnValue.Equals(value) Then
                                            ' If the column name does exist within the tempalte but the value associated with
                                            ' the column name does not equal what is in the tempplate then the test is 
                                            ' considered to have failed.
                                            color = "red"
                                            testPass = False
                                        End If
                                        ' Remove the CubeTuple for the current column name from the template data structure.
                                        currentCubeRow.RemoveCubeTuple(columnName)
                                    End If
                                    tableResult &= "<td white-space:nowrap bgcolor=""" & color & """>" & value & "</td>"
                                End If
                            End If
                        Next currentCol
                        tableResult &= "</tr>"
                    Next currentRow
				tableResult &= "</table>"

				If testPass Then
					' If the tests thus far has passed check to see if there is still CubeTuple(s) that exist within the
					' template.  If there are then extra data exists in the template that was not returned by the current
					' MDX query.
					testPass = Not CheckForExtraDataInTemplate(templateCubeDict)
				End If
				Return New CubeComparisonResult(testPass, tableResult)
			End Using
			End Using
		End Function

'''         Recursively checks the MDX query result dictionary for extra data that was not found
'''         * during a subsequent run.  What this means is that the test fails because it does not have as much 
'''         * data returned as the origional query.
'''         *
		Private Function CheckForExtraDataInTemplate(ByVal templateCubeDict As Dictionary(Of String, CubeRow)) As Boolean
			For Each key As String In templateCubeDict.Keys
				Dim cubeRow As CubeRow = templateCubeDict(key)
                If cubeRow.GetCubeDict IsNot Nothing Then
                    Return CheckForExtraDataInTemplate(cubeRow.GetCubeDict)
                Else
                    If cubeRow.GetCubeTupleSize > 0 Then
                        Return True
                    End If
                End If
			Next key
			Return False
		End Function

'''         Builds the HTML and Dictionary<string, CubeRow> MDX query result objects give an MDX query 
'''         * mdxQuery - The MDX query that contains the test
'''         * templateCubeDict - The datastructure that holds the MDX query results
'''         
		Private Function BuildTemplate(ByVal mdxQuery As String, ByVal templateCubeDict As Dictionary(Of String, CubeRow)) As String
			Dim tableResult As String = "<table border=""1"">"
			' Open a connection to the Analysis server
			Using connection = GetAdomdConnection(connectionstring)
			' Execute the MDX query
			Using command = New AdomdCommand(mdxQuery, connection)
				' Get the MDX query results as a CellSet object
                    Dim cst As CellSet = command.ExecuteCellSet
				' Make sure that the result contains at least two axes.  Columns and rows.
				Dim axesCount As Integer = cst.Axes.Count
				If axesCount <> 2 Then
					Throw New Exception("The MDX Query did not return two axes.  Please check your query.")
				End If

				' Make sure that data is contained within the two axes.
                    For count = 0 To axesCount - 1
                        If cst.Axes(count).Positions.Count = 0 Then
                            Throw New Exception("The MDX query return no data.  Please check your query.")
                        End If
                    Next count

				' Get the number of column dimensions
				Dim colDimCount As Integer = cst.Axes(0).Positions(0).Members.Count
				' Get the number of row dimensions
				Dim rowDimCount As Integer = 0
				If cst.Axes(1).Positions(0).Members.Count > 0 Then
					rowDimCount = cst.Axes(1).Positions(0).Members.Count
				End If

				' The row count is the total number of Position(s) on the row axis plus the number of column dimensions
				' The reason the column dimension count is included is that the column headers themselves make up rows
				Dim rowCount As Integer = cst.Axes(1).Positions.Count + colDimCount
				' The column count is the total number of Position(s) on the column axis plus the number of row dimensions.
				' The reason the row dimension count is included is that the row headers themselves make up columns
				Dim colCount As Integer = cst.Axes(0).Positions.Count + rowDimCount

				' Create an array of string(s) that hold the column headers
				Dim columnHeader(cst.Axes(0).Positions.Count - 1) As String

				' Create a pointer to hold the current CubeRow object
				Dim currentCubeRow As CubeRow = Nothing
				' Create a point to hold the current CubeRow dictionary
				Dim currentCubeDict As Dictionary(Of String, CubeRow) = templateCubeDict

				' Loop through all the rows within the MDX result
				For currentRow = 0 To rowCount - 1
					' Create a new HTML table row
					tableResult &= "<tr>"

					' If the current row is greater than the number of column dimensions then 
					' processing of actual row data is occurring.  Set the CubeRow pointer to a new
					' CubeRow and set the CubeRow Dictionary pointer to the original Dictionary passed in.
					' This is also the head of the Dictionary structure.
					If currentRow >= colDimCount Then
                            currentCubeRow = New CubeRow()
						currentCubeDict = templateCubeDict
					End If

					' Loop through all the columns within the MDX result
                        For currentCol = 0 To colCount - 1
                            ' Check if the current row is dealing with column dimension captions
                            If currentRow < colDimCount Then
                                If currentCol < rowDimCount Then
                                    ' If the current column is has not reached the row dimension headers create a blank HTML cell
                                    ' These cells make up the uper left hand corner of the HTML table as place holders before the 
                                    ' HTML column headers and row headers meet.
                                    tableResult &= "<th> </th>"
                                Else
                                    ' A column header is being processed.  The column header may be many levels deep depending on
                                    ' how many dimensions are included in the column portion of the MDX query.  The location of the column header is 
                                    ' computed by taking the current column count minus the number of row dimensions since getting the total number of 
                                    ' columns involved adding the total number of row dimensions.

                                    ' Get the current column index which is the current column count - the number of row dimensions since
                                    ' the row dimensions captions are part of the columns.
                                    Dim columnIndex As Integer = currentCol - rowDimCount

                                    ' If the column header at the given location does not exist add it.
                                    If columnHeader(columnIndex) Is Nothing Then
                                        columnHeader(columnIndex) = cst.Axes(0).Positions(columnIndex).Members(currentRow).Caption
                                    Else
                                        ' Otherwise if the column header does exist append to it.
                                        columnHeader(columnIndex) &= ":" & cst.Axes(0).Positions(columnIndex).Members(currentRow).Caption
                                    End If

                                    ' Create an HTML table header cell to hold the column name and add it to the HTML template.
                                    tableResult &= "<th>" & cst.Axes(0).Positions(columnIndex).Members(currentRow).Caption & "</th>"
                                End If
                            Else ' If current row is greater then or equal to the column dimension count then processing of the MDX row data commences
                                ' If the current column count is less than the row count then a row header is being traversed.
                                If currentCol < rowDimCount Then
                                    If currentCol > 0 Then
                                        ' If the current column count is greater than 0 then a new row header is being processed.
                                        ' Create a new CubeRow Dictionary to store the contents of the given row header.
                                        If currentCubeRow.GetCubeDict Is Nothing Then
                                            currentCubeRow.setCubeDict(New Dictionary(Of String, CubeRow))
                                        End If
                                        currentCubeDict = currentCubeRow.GetCubeDict
                                        currentCubeRow = New CubeRow
                                    End If

                                    ' Get the current row index which is the current row count - the number of column dimensions since the 
                                    ' column dimension captions are part of the rows.
                                    Dim rowIndex As Integer = currentRow - colDimCount
                                    ' Store the row header caption
                                    Dim rowCaption As String = cst.Axes(1).Positions(rowIndex).Members(currentCol).Caption
                                    If Not currentCubeDict.ContainsKey(rowCaption) Then
                                        ' If the row header caption does not exist in the current CubeRow dictionary add it and the 
                                        ' new CubeRow
                                        currentCubeDict.Add(rowCaption, currentCubeRow)
                                    Else
                                        ' if the row header does exist in the current CubeRow dictionary then use it to retrieve the current
                                        ' CubeRow value.
                                        currentCubeRow = currentCubeDict(rowCaption)
                                    End If

                                    ' Update the HTML to include the row header caption
                                    tableResult &= "<th>" & rowCaption & "</th>"
                                Else
                                    ' Get the current column index which is the current column count - the number of row dimensions since
                                    ' the row dimensions captions are part of the columns.
                                    Dim columnIndex As Integer = currentCol - rowDimCount
                                    ' Get the current row index which is the current row count - the number of column dimensions since the 
                                    ' column dimension captions are part of the rows.
                                    Dim rowIndex As Integer = currentRow - colDimCount
                                    ' Add a CubeTuple at the current column and row location with the current column header and Cube value
                                    currentCubeRow.AddTuple(columnHeader(columnIndex), cst(columnIndex, rowIndex).FormattedValue)
                                    ' Add the Cube cell value within an HTML cell.
                                    tableResult &= "<th>" & cst.Axes(0).Positions(columnIndex).Members(currentRow).Caption & "</th>"
                                End If
                            End If
                        Next currentCol
					tableResult &= "</tr>"
				Next currentRow
				tableResult &= "</table>"

				Return tableResult
			End Using
			End Using
		End Function

		''' Check if the MDX query result template with the given MDX test name exists *
		Private Function TemplateExists(ByVal mdxTestTitle As String) As Boolean
			Dim templateDir As String = GetTemplateDir(mdxTestTitle)
			Return System.IO.Directory.Exists(templateDir)
		End Function

		''' Returns the template directory for the given MDX test name *
		Private Function GetTemplateDir(ByVal mdxTestTitle As String) As String
			Return baseOutputDir & "\templates\" & mdxTestTitle
		End Function
	End Class
End Namespace
